/**
 * This class implements a ViewEventListener that is listening for the
 * specified
 */
class ViewEventListener {
  constructor(type, listenerFunction, args=[]) {
    this.listenerFunction = listenerFunction;
    this.type = type;
    this.args = args;
  }
}

/**
 * This class is implements a View which handles the reprensentation of the
 * data that has been fetched by the Client of the corpus_analysis. This view
 * only handles how the data is shown to the user. View extends the list.js
 * List class.
 */
class ResultsList extends List {
  /**
   * If no options are given when a new instance of this class is created
   * the options below are used.
   */
  static options = {
    page: 30,
    pagination: [{
      name: "paginationTop",
      paginationClass: "paginationTop",
      innerWindow: 8,
      outerWindow: 1
    }, {
      paginationClass: "paginationBottom",
      innerWindow: 8,
      outerWindow: 1
    }],
    valueNames: ["titles", "lc", "c", "rc", {data: ["index"]}],
    item: `<span></span>`
  };
  constructor(idOrElement, options) {
    super(idOrElement, options);
    /**
     * All span tokens which are holding events if expert
     * mode is on. Collected here to delete later on.
     */
    this.eventTokens = {};
    /**
     * All token elements which have added classes like chip and hoverable for
     * expert view. Collected here to delete later on.
     */
    this.currentExpertTokenElements = {};
    /**
     * Holds True/false for check buttons used to add matches to sub-results.
     * If checked, it is True. If unchecked, it is false. Buttons for this
     * have the class add. The ittle round check buttons to add matches to sub
     * results. Key is match index. Value is true or false as mentioned above.
     */
    this.subResultsIndexes = {};
    // ViewEventListeners listening for client notifications.
    this.notificationListeners = {};
    this.knownHTMLElements = new Set();
  }
  /**
   * Function to clear/reset some class field values. Usefull if a new query
   * hase been issued by the user.
   */
  resetFields() {
    this.subResultsIndexes = {};
  }


  /**
   * Function that takes one or more query selector
   * strings in an array as an input. The function then creates a
   * class field in the ResultsList object with the query selector
   * string as the key. The selector will be converted to a valid JavaScript
   * Field name i. e. #html-id-string -> this.htmlIdString
   * The value will be the identifed element or elements fetched with the
   * querySelector respectively querySelectorAll method.
   * If the query selector is passed as an Array of length 2, with the second
   * element defining modal options, teh identified element will be initialized
   * as a modal with the given options.
   */
  getHTMLElements(arrayOfSelectors) {
    for (let selector of arrayOfSelectors) {
      // Check if identified Element should be initialized as a modal.
      let modalInit = false;
      let options;
      if (Array.isArray(selector)) {
        options = selector[1];
        selector = selector[0];
        modalInit = true;
      }
      // Check if the current selector has already been used.
      if (this.knownHTMLElements.has(selector)) {
        continue;
      } else {
        // Get element or elements.
        let element;
        let elements;
        if (selector.startsWith('#')) {
          element = document.querySelector(selector);
        } else {
          elements = document.querySelectorAll(selector);
          elements = [...elements];
        }
        // Create valid javascript instance field name.
        let cleanKey = [];
        selector = selector.replace(/_/g, '-');
        selector.match(/\w+/g).forEach((word) => {
          let tmp = word[0].toUpperCase() + word.slice(1);
          cleanKey.push(tmp);
        });
        cleanKey[0] = cleanKey[0].toLowerCase();
        cleanKey = cleanKey.join('');
        this[cleanKey] = element ? element: elements;
        // Initialize current element as modal if modalInit true.
        if (modalInit) {
          this[cleanKey] = M.Modal.init(this[cleanKey], options);
        }
      }
      // Add current selector to knwonHTMLElements.
      this.knownHTMLElements.add(selector);
    }
  }

  /**
   * Register ViewEventListeners to the ResultsList. Which will listen for
   * the specified event.
   */
  setViewEventListeners(notificationListeners) {
    for (let notificationListener of notificationListeners) {
      this.notificationListeners[notificationListener.type] = notificationListener;
    }
  }

  /**
   * Loads the ViewEventListeners so that hey will be listening to their
   * assigned custom events.
   */
  loadViewEventListeners() {
    for (let [type, listener] of Object.entries(this.notificationListeners)) {
      if (listener.args.length > 0) {
        listener.listenerFunction(...listener.args);
      } else {
        listener.listenerFunction(type, this);
      }
    }
  }

  /**
   * This functions sends events to the Client to trigger specific functions to
   * trigger new data requests from the server.
   */
  notifyClient(caseIdentifier, detailObject={}) {
    detailObject.caseIdentifier = caseIdentifier;
    const event = new CustomEvent('notify-client', { detail: detailObject });
    console.info('Client dispatching Notification:', caseIdentifier);
    document.dispatchEvent(event);
  }

  /**
   * Creates cpos either from ranges or not.
   */
  helperCreateCpos(cpos_ranges, cpos_values) {
    let lc;
    let c;
    let rc;
    if (cpos_ranges) {
      /**
       * Python range like function from MDN
       * https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Array/from#Sequence_generator_(range)
       */
      const range = (start, stop, step) => Array.from({ length: (stop - start) / step + 1}, (_, i) => start + (i * step));
      lc = range(cpos_values.lc[0], cpos_values.lc[1], 1)
      c = range(cpos_values.c[0], cpos_values.c[1], 1)
      rc = range(cpos_values.rc[0], cpos_values.rc[1], 1)
    } else {
      lc = cpos_values.lc;
      c = cpos_values.c;
      rc = cpos_values.rc;
    }
    return {lc: lc, c: c, rc: rc};
  }

  // Get display options from display options form element.
  static getDisplayOptions(htmlId) {
    // gets display options parameters
    let displayOptionsFormElement = document.getElementById(htmlId);
    let displayOptionsFormData = new FormData(displayOptionsFormElement);
    let displayOptionsData =
      {
        "resultsPerPage": displayOptionsFormData.get("display-options-form-results_per_page"),
        "resultsContex": displayOptionsFormData.get("display-options-form-result_context"),
        "expertMode": displayOptionsFormData.get("display-options-form-expert_mode")
      };
    return displayOptionsData
  }
  /**
   * Used in addToSubResults and inspect to toggle the design of the check
   * buttons according to its checked unchecked status.
   */
  helperActivateAddBtn(btn) {
    btn.classList.remove("corpus-analysis-color.lighten");
    btn.classList.add("green");
    btn.textContent = "check";
  }

  /**
   * Used in addToSubResults and inspect to toggle the design of the check
   * buttons according to its checked unchecked status.
   */
  helperDeactivateAddBtn(btn) {
    btn.classList.remove("green");
    btn.classList.add("corpus-analysis-color.lighten");
    btn.textContent = "add";
  }
  /**
   * This function is invoked when the users adds or removes a match using the
   * add-btn (+ button/or green checkmark) to/from sub-results. When the button
   * is clicked the function checks if the current dataIndex ID is already
   * saved in subResultsIndexes or not. If it is not the dataIndex will be used
   * as a key in subResultsIndexes with the value true. If it is already added
   * the entry with the key dataIndex will be deleted from subResultsIndexes.
   * Visual feedback (green checkmark if a match has been added etc.) is also
   * handled on the basis of the information stored in subResultsIndexes.
   */
  addToSubResults(dataIndex, client, tableCall=true) {
    let toShowArray;
    dataIndex = parseInt(dataIndex);
    if (!this.subResultsIndexes[dataIndex]
        || this.subResultsIndexes[dataIndex] === undefined) {
      // add button is activated because status is false or undefined
      this.helperActivateAddBtn(event.target);
      this.subResultsIndexes[dataIndex] = true;
      toShowArray = Object.keys(this.subResultsIndexes).map(index => parseInt(index));
      // Add 1 because indexes are zero based. User sees 1 based numbering.
      toShowArray = toShowArray.map(index => index + 1);
      // Allways sort the shown indexes for the user if new match is added.
      toShowArray = toShowArray.sort(function(a, b){return a-b});
      this.subResultsIndexesDisplay.textContent = toShowArray.join(', ');
      M.textareaAutoResize(this.subResultsIndexesDisplay);
      this.nrMarkedMatches.textContent = Object.keys(this.subResultsIndexes).length;
    } else if (this.subResultsIndexes[dataIndex]) {
      // add button is deactivated because status is true
      this.helperDeactivateAddBtn(event.target);
      delete this.subResultsIndexes[dataIndex];
      toShowArray = Object.keys(this.subResultsIndexes).map(index => parseInt(index));
      // Add 1 because indexes are zero based. User sees 1 based numbering.
      toShowArray = toShowArray.map(index => index + 1);
      // Allways sort the shown indexes for the user if new match is added.
      toShowArray = toShowArray.sort(function(a, b){return a-b});
      this.subResultsIndexesDisplay.textContent = toShowArray.join(', ');
      this.nrMarkedMatches.textContent = Object.keys(this.subResultsIndexes).length;
      M.textareaAutoResize(this.subResultsIndexesDisplay);
    }
    // Toggles create button according to the number of ids in addToSubResultsIdsToShow
    if (Object.keys(this.subResultsIndexes).length > 0 && !client.isBusy) {
      this.subResultsCreate.classList.toggle('disabled', false);
    } else if (Object.keys(this.subResultsIndexes).length === 0) {
      this.subResultsCreate.classList.toggle('disabled', true);
    }
    /**
     * After a match as been added or removed the export button will be
     * hidden because the sub-results have been altered and have to be built
     * again. Thus subResultsCreateElement has to be shown again.
     */
    this.subResultsExport.classList.add("hide");
    this.subResultsCreate.classList.remove("hide");
    /**
     * Also activate/deactivate buttons in the table/resultsList accordingly
     * if button in inspect was activated/deactivated.
     * This part only runs if tableCall is set to false when this function is
     * called.
     */
    if (!tableCall) {
      this.getHTMLElements(['#query-results-table']);
      let container = this.queryResultsTable.querySelector(`[data-index="${dataIndex}"]`);
      let tableAddBtn = container.querySelector('.add-btn'); // gets the add button from the list view
      if (this.subResultsIndexes[dataIndex]) {
        this.helperActivateAddBtn(tableAddBtn);
      } else {
        this.helperDeactivateAddBtn(tableAddBtn);
      }
    }
  }

  // Toggle inspect buttons depending on the Client status
  toggleInspectButtons(client) {
    if (!client.isBusy) {
      this.activateInspectButtons();
    } else if (client.isBusy) {
      this.deactivateInspectButtons();
    }
  }

  // Helper function. Should be private if feature is available.
  activateInspectButtons() {
    let inspectBtnElements;
    inspectBtnElements = document.querySelectorAll('.inspect');
    for (let inspectBtn of inspectBtnElements) {
      inspectBtn.classList.toggle('disabled', false);
    }
  }

  // Helper function. Should be private if feature is available.
  deactivateInspectButtons() {
    let inspectBtnElements;
    inspectBtnElements = document.querySelectorAll('.inspect');
    for (let inspectBtn of inspectBtnElements) {
      inspectBtn.classList.toggle('disabled', true);
    }
  }
  /**
   * Function to inspect one match in more detail (Showing more context).
   * If in dynamic mode the view notifies the client to requests the new
   * context for the one match identified by the given dataIndex.
   * If not in dynamic mode the the needed context will be gathered from the
   * already present results in results.data.
   */
  inspect(client, results, dataIndex, type) {
    // initialize context modal
    this.getHTMLElements([
      ['#context-modal', true],
      '#context-results',
      '#create-inspect-menu',
      '#create-from-inspect',
    ]);
    // Clear fields from old data on every new inspect() call.
    this.contextId = dataIndex[0];
    this.contextResults.innerHTML = '';
    // Open modal.
    this.contextModal.open();
    this.contextResults.insertAdjacentHTML('afterbegin', `
      <div class="progress">
        <div class="indeterminate"></div>
      </div>
    `);
    if (client.dynamicMode) {
      // Notify Client to get results from server.
      this.notifyClient('get-results', {resultsType: 'inspect-results',
                                        dataIndexes: dataIndex,
                                        resultsList: this});
    } else {
      // Gather results data from already present data.
      results.inspectResultsData.matches = [results.data.matches[dataIndex[0]]];
      results.inspectResultsData.cpos_ranges = results.data.cpos_ranges;
      this.showMatchContext(results, client)
    }
    // Match nr for user to display derived from data_index.
    let contextMatchNrElement = document.getElementById("context-match-nr");
    contextMatchNrElement.textContent = this.contextId + 1;
    // Add the add button to add this match to sub results with onclick event.
    let classes = `btn-floating btn waves-effect` +
                  ` waves-light corpus-analysis-color.lighten right`
    let addToSubResultsIdsBtn = document.createElement("a");
    addToSubResultsIdsBtn.setAttribute("class", classes + ` add`);
    addToSubResultsIdsBtn.innerHTML = '<i class="material-icons">add</i>';
    addToSubResultsIdsBtn.onclick= () => {
        this.addToSubResults(dataIndex[0], client, false)
      };
    /**
     * Checks if the match has or has not been added to sub results yet.
     * Sets the color and status of the button accordingly.
     */
    if (this.subResultsIndexes[dataIndex[0]]) {
      this.helperActivateAddBtn(addToSubResultsIdsBtn.firstElementChild);
    } else if (!this.subResultsIndexes[dataIndex[0]]) {
      this.helperDeactivateAddBtn(addToSubResultsIdsBtn.firstElementChild);
    }
    this.createInspectMenu.innerHTML = '';
    this.createInspectMenu.appendChild(addToSubResultsIdsBtn);
    // Hide create menu if not in dynamic mode.
    if (!client.dynamicMode) {
      this.createFromInspect.classList.add('hide');
    }
  }
  /**
   * Create Element from HTML String. Helper function should be private.
   * https://stackoverflow.com/questions/494143/creating-a-new-dom-element-from-an-html-string-using-built-in-dom-methods-or-pro/35385518#35385518
   */
  HTMLTStrToElement(htmlStr) {
    let template = document.createElement("template");
    htmlStr = htmlStr.trim();
    template.innerHTML = htmlStr;
    return template.content.firstChild;
    }
  /**
   * Used either as a callback if the client has been notified to get new
   * results with new full context. Or just directly invoced as a function
   * with the according input data.
   */
  showMatchContext(results, client) {
    this.getHTMLElements([
      '#context-results',
      '#inspect-display-options-form-expert_mode_inspect',
      '#inspect-display-options-form-highlight_sentences',
      '#context-sentences'
    ])
    let uniqueS = new Set();
    let uniqueContextS = new Set();
    let {lc, c, rc} = this.helperCreateCpos(results.inspectResultsData.cpos_ranges,
                                            results.inspectResultsData.matches[0]);
    // Create sentence strings as tokens.
    let tokenHTMLArray = [];
    let htmlTokenStr = ``;
    let tokenHTMlElement;
    let token;
    for (let cpos of lc) {
      if (client.dynamicMode) {
        token = results.inspectResultsData.cpos_lookup[cpos];
      // If client is not in dynamic mode use cpos_lookup from results.data
      } else {
        token = results.data.cpos_lookup[cpos];
      }
      uniqueS.add(token.s)
      htmlTokenStr = `<span class="token"` +
                           `data-sid="${token.s}"` +
                           `data-cpos="${cpos}">` +
                       `${token.word}` +
                     `</span>`;
      tokenHTMlElement = this.HTMLTStrToElement(htmlTokenStr)
      tokenHTMLArray.push(tokenHTMlElement);
    }
    for (let cpos of c) {
      if (client.dynamicMode) {
        token = results.inspectResultsData.cpos_lookup[cpos];
      // If client is not in dynamic mode use cpos_lookup from results.data
      } else {
        token = results.data.cpos_lookup[cpos];
      }
      uniqueContextS.add(token.s);
      uniqueS.add(token.s);
      htmlTokenStr = `<span class="token bold light-green"` +
                           `data-sid="${token.s}"` +
                           `data-cpos="${cpos}"` +
                           `style="text-decoration-line: underline;">` +
                       `${token.word}` +
                     `</span>`;
      tokenHTMlElement = this.HTMLTStrToElement(htmlTokenStr)
      tokenHTMLArray.push(tokenHTMlElement);
    }
    results.inspectResultsData["context_s_ids"] = Array.from(uniqueContextS);
    for (let cpos of rc) {
      if (client.dynamicMode) {
        token = results.inspectResultsData.cpos_lookup[cpos];
      // If client is not in dynamic mode use cpos_lookup from results.data
      } else {
        token = results.data.cpos_lookup[cpos];
      }
      uniqueS.add(token.s)
      htmlTokenStr = `<span class="token"` +
                           `data-sid="${token.s}"` +
                           `data-cpos="${cpos}">` +
                       `${token.word}` +
                     `</span>`;
      tokenHTMlElement = this.HTMLTStrToElement(htmlTokenStr)
      tokenHTMLArray.push(tokenHTMlElement);
    }
    // Remove loading indeterminate HTML before context is inserted.
    this.contextResults.innerHTML = '';
    for (let sId of uniqueS) {
      let htmlSentence = `<span class="sentence" data-sid="${sId}"></span>`;
      let sentenceElement = this.HTMLTStrToElement(htmlSentence);
      for (let tokenElement of tokenHTMLArray) {
        if (tokenElement.dataset.sid == sId) {
          sentenceElement.appendChild(tokenElement);
          sentenceElement.insertAdjacentHTML("beforeend", ` `);
        } else {
          continue;
        }
      }
      this.contextResults.appendChild(sentenceElement);
    }
    // Add expert mode switch event for the modal to toggle expert mode.
    this.inspectDisplayOptionsFormExpertModeInspect.onchange = (event) => {
      if (event.target.checked) {
        this.expertModeOn("context-results", results);
      } else {
        this.expertModeOff("context-results")
      }
    };
    // Add switch event to toggle Sentence highlighting.
    this.inspectDisplayOptionsFormHighlightSentences.onchange = (event) => {
      if (event.target.checked) {
        this.higlightContextSentences();
      } else {
      this.unhighlightContextSentences();
      }
    };
    // Add range event to change nr of context sentences.
    this.contextSentences.onchange = (event) => {
      this.changeSentenceContext(event.target.value);
    }
    /**
     * Checks on new modal opening if switches are checked
     * if switches are checked functions are executed.
     */
    if (this.inspectDisplayOptionsFormExpertModeInspect.checked) {
      this.expertModeOn("context-results", results);
    }
    if (this.inspectDisplayOptionsFormHighlightSentences.checked) {
      this.higlightContextSentences();
    }
    /**
     * Checks the value of the number of sentences to show on modal opening
     * sets context sentences accordingly
     */
    this.changeSentenceContext(this.contextSentences.value);
  }

  // Splits context text into sentences based on spacy sentence split
  higlightContextSentences() {
    let sentences = document.getElementById("context-results").getElementsByClassName("sentence");
    for (let s of sentences) {
      s.insertAdjacentHTML("beforeend", `<span><br><br></span>`)
    }
  }
  // Reverse operation of above function.
  unhighlightContextSentences() {
    let sentences = document.getElementById("context-results").getElementsByClassName("sentence");
    let br;
    for (let s of sentences) {
      br = s.lastChild;
      br.remove();
    }
  }

  // Changes how many context sentences in inspect view are shown.
  changeSentenceContext(sValue, maxSValue=10) {
    sValue = maxSValue - sValue;
    // console.log(sValue);
    let sentences = document.getElementById("context-results").getElementsByClassName("sentence");
    let array = Array.from(sentences);
    let toHideArray;
    let toShowArray;
    if (sValue != 0) {
      toHideArray = array.slice(0, sValue).concat(array.slice(-(sValue)));
      toShowArray = array.slice(sValue, 9).concat(array.slice(9, -(sValue)))
    } else {
      toHideArray = [];
      toShowArray = array;
    }
    for (let s of toHideArray) {
      s.classList.add("hide");
    }
    for (let s of toShowArray) {
      s.classList.remove("hide");
    }
  }

  // ###### Display options functions changing how results are being displayed ######

  /**
   * Event function that changes the shown hits per page.
   * Just alters the resultsList.page property.
   */
  changeHitsPerPage(client, results) {
    this.page = this.displayOptionsFormResultsPerPage.value;
    this.update();
    this.changeContext();
    this.toggleInspectButtons(client);
    if (this.displayOptionsFormExpertMode.checked) {
      this.expertModeOn('query-display', results);
    }
  }
  /**
   * Event function triggered on context select change also if pagination is
   * clicked.
   */
  changeContext() {
    let newContextValue = this.displayOptionsFormResultContext.value;
    let lc = document.querySelectorAll(".left-context");
    let rc = document.querySelectorAll(".right-context");
    for (let element of lc) {
      let arrayLc = Array.from(element.childNodes);
      for (let element of arrayLc.reverse().slice(newContextValue)) {
        element.classList.add("hide");
      }
      for (let element of arrayLc.slice(0, newContextValue)) {
        element.classList.remove("hide");
      }
    }
    for (let element of rc) {
      let arrayRc = Array.from(element.childNodes);
      for (let element of arrayRc.slice(newContextValue)) {
        element.classList.add("hide");
      }
      for (let element of arrayRc.slice(0, newContextValue)) {
        element.classList.remove("hide");
      }
    }
  }

  // ###### Expert view event functions ######

  // Function to create a tooltip for the current hovered token.
  tooltipEventCreate(event, results) {
    let token = results.data.cpos_lookup[event.target.dataset.cpos];
    if (!token) {
      token = results.inspectResultsData.cpos_lookup[event.target.dataset.cpos];
    }
    this.currentTooltipElement = M.Tooltip.init(event.target, {
      'html': `<table>
                 <tr>
                   <th>Token information</th>
                   <th>Source information</th>
                 </tr>
                 <tr>
                   <td class="left-align">
                     Word: ${token.word}<br>
                     Lemma: ${token.lemma}<br>
                     POS: ${token.pos}<br>
                     Simple POS: ${token.simple_pos}<br>
                     NER: ${token.ner}
                   </td>
                   <td class="left-align">
                     Title: ${results.data.text_lookup[token.text].title}
                     <br>
                     Author: ${results.data.text_lookup[token.text].author}
                     <br>
                     Publishing year: ${results.data.text_lookup[token.text].publishing_year}
                   </td>
                 </tr>
               </table>`
      }
    );
  }
  /**
  * Function to destroy the current Tooltip for the current hovered tooltip
  * on mouse leave
  */
  tooltipEventDestroy(event) {
    this.currentTooltipElement.destroy();
  }

  /**
  * Turn the expert mode on for all tokens in the DOM element identified by
  * its htmlID.
  */
  expertModeOn(htmlId, results) {
    if (!Array.isArray(this.currentExpertTokenElements[htmlId])) {
      this.currentExpertTokenElements[htmlId] = [];
    }
    let container = document.getElementById(htmlId);
    let tokens = container.querySelectorAll('span.token');
    this.currentExpertTokenElements[htmlId].push(...tokens);
    this.eventTokens[htmlId] = [];
    for (let tokenElement of this.currentExpertTokenElements[htmlId]) {
      tokenElement.classList.add('chip', 'hoverable', 'expert-view');
      const eventCreate = (event, arg) => this.tooltipEventCreate(event, arg);
      tokenElement.onmouseover = (event) => eventCreate(event, results);
      tokenElement.onmouseout = (event) => this.tooltipEventDestroy(event);
      this.eventTokens[htmlId].push(tokenElement);
    }
  }

  /**
  * Turn the expert mode off for all tokens in the DOM element identified by
  * its htmlID.
  */
  expertModeOff(htmlId) {
    if (!Array.isArray(this.currentExpertTokenElements[htmlId])) {
      this.currentExpertTokenElements[htmlId] = [];
    }
    if (!Array.isArray(this.eventTokens[htmlId])) {
      this.eventTokens[htmlId] = [];
    }
    for (let tokenElement of this.currentExpertTokenElements[htmlId]) {
      tokenElement.classList.remove("chip", "hoverable", "expert-view");
    }
    this.currentExpertTokenElements[htmlId] = [];

    for (let eventToken of this.eventTokens[htmlId]) {
      eventToken.onmouseover = '';
      eventToken.onmouseout = '';
    }
  this.eventTokens[htmlId] = [];
  }

  createResultRowElement(item, chunk, client, imported=false) {
    // Gather values from item.
    let values = item.values();
    let {lc, c, rc} = this.helperCreateCpos(chunk.cpos_ranges,
                                            values)
    // Get infos for full match row.
    let matchRowElement = document.createElement("tr");
    matchRowElement.setAttribute("data-index", values.index)
    let lcCellElement = document.createElement("td");
    lcCellElement.classList.add("left-context");
    matchRowElement.appendChild(lcCellElement);
    for (let cpos of lc) {
      let token = chunk.cpos_lookup[cpos];
      lcCellElement.insertAdjacentHTML("beforeend",
        `<span class="token" data-cpos="${cpos}">${token.word} </span>`);
    }

    // Get infos for hit of match and set actions.
    let textTitles = new Set();
    let aCellElement = document.createElement("td");
    aCellElement.classList.add("actions");
    let cCellElement = document.createElement("td");
    cCellElement.classList.add("match-hit");
    let textTitlesCellElement = document.createElement("td");
    textTitlesCellElement.classList.add("titles");
    let matchNrElement = document.createElement("td");
    matchNrElement.classList.add("match-nr");
    matchRowElement.appendChild(cCellElement);
    matchRowElement.appendChild(aCellElement);
    for (let cpos of c) {
      let token = chunk.cpos_lookup[cpos];
      cCellElement.insertAdjacentHTML("beforeend",
        `<span class="token" data-cpos="${cpos}">${token.word} </span>`);
      // Get text titles of every hit cpos token.
      textTitles.add(chunk.text_lookup[token.text].title);
    }
    // Add some interaction buttons.
    let css = `margin-right: 5px; margin-bottom: 5px;`
    let classes = `btn-floating btn waves-effect` +
                  ` waves-light corpus-analysis-color.lighten`
    // Add inspect button to trigger inspect view with more context.
    let inspectBtn = document.createElement("a");
    inspectBtn.setAttribute("style", css);
    inspectBtn.setAttribute("class", classes + ` disabled inspect`
                            );
    inspectBtn.innerHTML = '<i class="material-icons inspect-btn">search</i>';
    // Add btn to be able add matches to sub-results.
    let addToSubResultsBtn = document.createElement("a");
    addToSubResultsBtn.setAttribute("style", css);
    addToSubResultsBtn.setAttribute("class", classes + ` add`);
    addToSubResultsBtn.innerHTML = '<i class="material-icons add-btn">add</i>';
    if (client.dynamicMode || client.fullContext) {
      aCellElement.appendChild(inspectBtn);
    }
    if (client.dynamicMode) {
      aCellElement.appendChild(addToSubResultsBtn);
    }
    // Add text titles at front as first td of one row.
    textTitlesCellElement.textContent = [...textTitles].join(", ");
    matchRowElement.insertAdjacentHTML("afterbegin", textTitlesCellElement.outerHTML);
    matchNrElement.textContent = values.index + 1;
    matchRowElement.insertAdjacentHTML("afterbegin", matchNrElement.outerHTML);

    // Get infos for right context of match
    let rcCellElement = document.createElement("td");
    rcCellElement.classList.add("right-context");
    matchRowElement.appendChild(rcCellElement);
    for (let cpos of rc) {
      let token = chunk.cpos_lookup[cpos];
      rcCellElement.insertAdjacentHTML("beforeend",
        `<span class="token" data-cpos="${cpos}">${token.word} </span>`);
    }
    return matchRowElement
  }
  /**
   * Creates the HTML table code for the metadata view in the corpus analysis
   * interface
   */
  createMetaDataForModal(metaDataObject) {
    let html = `<div class="col s12">
                      <table class="highlight">
                        <thead>
                          <tr>
                            <th>Meta Data Description</th>
                            <th>Value</th>
                          </tr>
                        </thead>
                        <tbody>`
    for (let [outerKey, outerValue] of Object.entries(metaDataObject)) {
      // Use more descriptive names.
      if (outerKey === 'corpus_all_texts') {
        let tmpName = 'All texts in this corpus';
        html += `<tr>
                   <td style="text-transform: uppercase;">${tmpName.replace(/_/g, " ")}
                 </td>`
      } else if (outerKey === 'text_lookup') {
        let tmpName = 'Texts where the query resulted in matches'
        html += `<tr>
                   <td style="text-transform: uppercase;">${tmpName.replace(/_/g, " ")}
                 </td>`
      } else {
        html += `<tr>
                   <td style="text-transform: uppercase;">${outerKey.replace(/_/g, " ")}
                 </td>`
      }
      if (outerKey === "corpus_all_texts" || outerKey === "text_lookup") {
        html += `<td>
                  <ul class="collapsible">`
        for (let [innerKey, innerValue] of Object.entries(outerValue)) {
          html += `<li class="text-metadata"
                        data-metadata-key="${outerKey}"
                        data-text-key="${innerKey}">
                      <div class="collapsible-header"
                           data-metadata-key="${outerKey}"
                           data-text-key="${innerKey}">
                        <i class="material-icons"
                           data-metadata-key="${outerKey}"
                           data-text-key="${innerKey}">info_outline</i>
                           ${innerValue['author']} - ${innerValue['publishing_year']} -
                           ${innerValue['title']}
                      </div>
                      <div class="collapsible-body">
                        <span>
                          <ul id="bibliographic-data-${outerKey}-${innerKey}"
                              style="column-count: 2;">
                          </ul>
                        </span>
                      </div>
                    </li>`
        }
        html += `</ul>
                  </td>`
      } else {
        html += `<td>${outerValue}</td>`
      }
      html += `</tr>`
    }
    html += `</tbody>
              </table>`
    return html
  }
  /**
   * Creates the text details for the texts shown in the corpus analysis
   * metadata modal table.
   */
  createTextDetails(metaData) {
    let metadataKey = event.target.dataset.metadataKey;
    let textKey = event.target.dataset.textKey;
    let textData = metaData[metadataKey][textKey];
    let bibliographicData = document.querySelector(`#bibliographic-data-${metadataKey}-${textKey}`);
    bibliographicData.textContent = '';
    for (let [key, value] of Object.entries(textData)) {
      bibliographicData.insertAdjacentHTML("afterbegin",
        `
          <li>
            <span style="text-transform: capitalize;">${key}:</span>
            ${value}
          </li>
        `
      );
    }
  }
};

// Export classes.
export {
  ViewEventListener,
  ResultsList
};